/*$Id: shared.c,v 1.16 2002/12/09 10:37:35 aa5779 Exp $*/
/*
	Copyright (C) 2001, 2002 Artem V. Andreev (artem@AA5779.spb.edu)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <string.h>
#include <gfnmatch.h>
#include "lists.h"
#include "util.h"
#include "dstring.h"
#include "interp.h"

#define SI_HASH_SIZE 2048

/** The hash function is taken from GNU ObjectiveC runtime library;
namely, from objc/hash.h
*/
static unsigned int hash_string(const char *key)
{
	unsigned int ret = 0;
	unsigned int ctr = 0;
        
	while(*(char*)key) 
	{
		ret ^= *(char*)key++ << ctr;
		ctr = (ctr + 1) % sizeof(void *);
	}
	return ret % SI_HASH_SIZE;
}

DEFLIST(shared_int_hash,
		shared_int *si;
	);

typedef struct si_hash_key
{
	shared_object_class *cls;
	void *env;
	char *name;
} si_hash_key;

static shared_int_hash *si_hash_table[SI_HASH_SIZE];

extern shared_object_class *get_so_class(void *cls_name);

static char *get_class_name(void *cls)
{
	return get_so_class(cls)->id;
}

int find_byref(void *obj, void *key)
{
	return obj == key;
}

static int find_in_names(exported_names *en, char *key)
{
	return !strcmp(en->extname, key);
}

static void destroy_extname(exported_names *en)
{
	free(en->extname);
}

static int si_find_hash(shared_int_hash *hash, shared_int *si)
{
	return hash->si == si;
}

static int si_find_hash_bykey(shared_int_hash *hash, si_hash_key *key)
{
	return hash->si->so->class == key->cls && hash->si->env == key->env && !strcmp(hash->si->intname, key->name);
}

int so_adjust;
int so_find(void *obj, char *name)
{
	shared_object *so = (shared_object *)((char *)obj + so_adjust);
	return list_find(so->extnames, name, (list_predicate_t)find_in_names) != NULL;
}

int si_find_exact(shared_int *si, char *name)
{
	return !strcmp(si->intname, name);
}

int si_find_match(shared_int *si, char *key)
{
	return !fnmatch(si->intname, key, 0);
}

int si_find_byso(shared_int *si, void *so)
{
	return si->obj == so;
}

void destroy_shared_int(shared_int *si)
{
	shared_object *so = si->so;
	shared_object_class *soc = so->class;
	unsigned int hv = hash_string(si->intname);
	si_hash_table[hv] = list_remove(si_hash_table[hv], si, (list_predicate_t)si_find_hash, NULL, sizeof(shared_int_hash));
	free(si->intname);
	if(!--so->refcnt)
	{
		/* Use a tricky scheme instead of just list_remove to avoid problems with
			 list_remove called on the same base in soc->destroy
		 */
		list_t *to_remove = list_find(*soc->base, si->obj, (list_predicate_t)find_byref);
		if(!to_remove)
			fatal_error("shared object list corrupted");
		if(soc->destroy)
			soc->destroy(to_remove);
		*soc->base = list_remove(*soc->base, si->obj, (list_predicate_t)find_byref, NULL, soc->size);
	}
}

void destroy_so(shared_object *so)
{
	list_delete(so->extnames, (list_destroy_t)destroy_extname, sizeof(exported_names));
}

void export_shared(shared_object *so, char *extname)
{
	exported_names *en = ALLOC(exported_names);
	en->extname = xstrdup(extname);
	so->extnames = list_add(so->extnames, en);
}

void *find_exported(void *clsname, char *name)
{
	shared_object_class *cls = get_so_class(clsname);
	so_adjust = cls->offs;
	return list_find(*cls->base, name, (list_predicate_t)so_find);
}

shared_int *embed_shared(void *object, void *clsname, char *name)
{
	shared_object_class *cls = get_so_class(clsname);
 	shared_int *si;

	si = ALLOC(shared_int);
	si->intname = xstrdup(name);
	si->obj = object;
	si->so = (shared_object *)((char *)object + cls->offs);
	si->so->refcnt++;
	return si;
}

shared_int *import_shared(char *iname, char *name, void *clsname)
{
	shared_object_class *cls = get_so_class(clsname);
	void *so;
	so_adjust = cls->offs;
	so = list_find(*cls->base, name, (list_predicate_t)so_find);
	if(!so)
		recover_error("can't import %s '%s'", get_class_name(clsname), name);
	return embed_shared(so, clsname, iname);
}


shared_int *init_shared_object(void *env, void *so, void *clsname, char *intname)
{
	shared_object_class *soc = get_so_class(clsname);
	shared_object *so1;
	shared_int *si;
	
	so1 = (shared_object *)((char *)so + soc->offs);
	so1->refcnt = 1;
	so1->class = soc;
	so1->owner = env;
	*soc->base = list_add(*soc->base, so);
	si = ALLOC(shared_int);
	si->intname = xstrdup(intname);
	si->obj = so;
	si->env = NULL;
	si->so = so1;
	return si;
}

shared_int *clone_shared_object(void *env, void *so, void *clsname, char *intname)
{
	shared_object_class *soc = get_so_class(clsname);
	void *new_so;
	if(!soc->clone)
		recover_error("cannot clone a %s", soc->id);
	new_so = soc->clone(so);
	return init_shared_object(env, new_so, clsname, intname);
}

shared_int *get_si_list(void *env, void *cls)
{
	shared_object_class *soc = get_so_class(cls);
	return *(shared_int **)((char *)env + soc->offs2);
}

shared_int *find_si_itself(void *env, void *cls, char *iname)
{
	si_hash_key key;
	shared_int_hash *sh;
	key.cls = get_so_class(cls);
	key.env = env;
	key.name = iname;
	sh = list_find(si_hash_table[hash_string(iname)], &key, 
							 (list_predicate_t)si_find_hash_bykey);
	return sh ? sh->si : NULL;
}

void *quiet_find_shared_int(void *env, void *cls, char *iname)
{
	shared_int *si = find_si_itself(env, cls, iname);
	return si ? si->obj : NULL;
}

void *find_shared_int(void *env, void *cls, char *iname)
{
	void *so = quiet_find_shared_int(env, cls, iname);
	if(!so)
		recover_error("%s '%s' undefined", get_class_name(cls), iname);
	return so;
}

static int al_find(allow_list *al, void *key)
{
	return !al->subject || al->subject == key;
}

static int al_find_exact(allow_list *al, void *key)
{
	return al->subject == key;
}

void check_si_perm(void *env, shared_int *si)
{
	if(env != si->so->owner && !list_find(si->so->allowed, env, (list_predicate_t)al_find) &&
	   list_find(si->so->denied, env, (list_predicate_t)al_find))
		recover_error("permission denied to use an object");
}

void allow_shared_int(void *owner, shared_int *si, void *dest)
{
	allow_list *al;
	if(si->so->owner != owner)
		recover_error("only object owner may change its access lists");
	if(!dest)
	{
		list_delete(si->so->denied, NULL, sizeof(*si->so->denied));
		si->so->denied = NULL;
	}
	else
		si->so->denied = list_remove(si->so->denied, dest, (list_predicate_t)al_find_exact, NULL, sizeof(*si->so->denied));
	al = ALLOC(allow_list);
	al->subject = dest;
	si->so->allowed = list_add(si->so->allowed, al);
}

void deny_shared_int(void *owner, shared_int *si, void *dest)
{
	allow_list *al;
	if(si->so->owner != owner)
		recover_error("only object owner may change its access lists");
	if(!dest)
	{
		list_delete(si->so->allowed, NULL, sizeof(*si->so->allowed));
		si->so->allowed = NULL;
	}
	else
		si->so->allowed = list_remove(si->so->allowed, dest, (list_predicate_t)al_find_exact, NULL, sizeof(*si->so->denied));
	al = ALLOC(allow_list);
	al->subject = dest;
	si->so->denied = list_add(si->so->denied, al);
}

void rename_shared_int(void *env, void *cls, char *name, char *newname, int force)
{
	shared_int *si = find_si_itself(env, cls, name);
	unsigned int hv = hash_string(si->intname);
	shared_int_hash *sh;

	if(!si)
	{
		if(force)
			return;
		recover_error("can't rename %s %s: not defined", get_class_name(cls), name);
	}
	si_hash_table[hv] = list_remove(si_hash_table[hv], si, (list_predicate_t)si_find_hash, NULL, sizeof(shared_int_hash));
	free(si->intname);
	si->intname = xstrdup(newname);
	hv = hash_string(newname);
	sh = ALLOC(shared_int_hash);
	sh->si = si;
	si_hash_table[hv] = list_add(si_hash_table[hv], sh);
}

void add_shared_int(void *env, shared_int *si)
{
	shared_int **lst = (shared_int **)((char *)env + si->so->class->offs2);
	shared_int_hash *sh;
	unsigned int hv = hash_string(si->intname);

	check_si_perm(env, si);
	si->env = env;
	*lst = list_add(*lst, si);
	sh = ALLOC(shared_int_hash);
	sh->si = si;
	si_hash_table[hv] = list_add(si_hash_table[hv], sh);
}

void append_shared_int(void *env, shared_int *si)
{
	shared_int **lst = (shared_int **)((char *)env + si->so->class->offs2);
	shared_int_hash *sh;
	unsigned int hv = hash_string(si->intname);

	check_si_perm(env, si);
	si->env = env;
	*lst = list_append(*lst, si);
	sh = ALLOC(shared_int_hash);
	sh->si = si;
	si_hash_table[hv] = list_add(si_hash_table[hv], sh);
}

void remove_shared_int(void *env, void *clsname, char *name)
{
	shared_object_class *class = get_so_class(clsname);	
	shared_int **lst = (shared_int **)((char *)env + class->offs2);
	*lst = list_remove(*lst, name, (list_predicate_t)si_find_exact, 
					   (list_destroy_t)destroy_shared_int, sizeof(shared_int));
}

